package aoc2023

import CharGrid
import println
import readInput

/*
--- Day 13: Point of Incidence ---
With your help, the hot springs team locates an appropriate spring which launches you neatly and precisely up to the edge of Lava Island.
There's just one problem: you don't see any lava.
You do see a lot of ash and igneous rock; there are even what look like gray mountains scattered around. After a while, you make your way to a nearby cluster of mountains only to discover that the valley between them is completely full of large mirrors.  Most of the mirrors seem to be aligned in a consistent way; perhaps you should head in that direction?
As you move through the valley of mirrors, you find that several of them have fallen from the large metal frames keeping them in place. The mirrors are extremely flat and shiny, and many of the fallen mirrors have lodged into the ash at strange angles. Because the terrain is all one color, it's hard to tell where it's safe to walk or where you're about to run into a mirror.
You note down the patterns of ash (.) and rocks (#) that you see as you walk (your puzzle input); perhaps by carefully analyzing these patterns, you can figure out where the mirrors are!
For example:
#.##..##.
..#.##.#.
##......#
##......#
..#.##.#.
..##..##.
#.#.##.#.

#...##..#
#....#..#
..##..###
#####.##.
#####.##.
..##..###
#....#..#

To find the reflection in each pattern, you need to find a perfect reflection across either a horizontal line between two rows or across a vertical line between two columns.
In the first pattern, the reflection is across a vertical line between two columns; arrows on each of the two columns point at the line between the columns:
123456789
    ><
#.##..##.
..#.##.#.
##......#
##......#
..#.##.#.
..##..##.
#.#.##.#.
    ><
123456789

In this pattern, the line of reflection is the vertical line between columns 5 and 6. Because the vertical line is not perfectly in the middle of the pattern, part of the pattern (column 1) has nowhere to reflect onto and can be ignored; every other column has a reflected column within the pattern and must match exactly: column 2 matches column 9, column 3 matches 8, 4 matches 7, and 5 matches 6.
The second pattern reflects across a horizontal line instead:
1 #...##..# 1
2 #....#..# 2
3 ..##..### 3
4v#####.##.v4
5^#####.##.^5
6 ..##..### 6
7 #....#..# 7

This pattern reflects across the horizontal line between rows 4 and 5. Row 1 would reflect with a hypothetical row 8, but since that's not in the pattern, row 1 doesn't need to match anything. The remaining rows match: row 2 matches row 7, row 3 matches row 6, and row 4 matches row 5.
To summarize your pattern notes, add up the number of columns to the left of each vertical line of reflection; to that, also add 100 multiplied by the number of rows above each horizontal line of reflection. In the above example, the first pattern's vertical line has 5 columns to its left and the second pattern's horizontal line has 4 rows above it, a total of 405.
Find the line of reflection in each of the patterns in your notes. What number do you get after summarizing all of your notes?

*/
fun main() {

    val inlineTestInput = """
#.##..##.
..#.##.#.
##......#
##......#
..#.##.#.
..##..##.
#.#.##.#.

#...##..#
#....#..#
..##..###
#####.##.
#####.##.
..##..###
#....#..#
"""

    fun mirrored(bits: IntArray, size: Int): Int {
        for (p1 in 0 until size - 1) {
            var p2 = 0
            var match = true
            while (p1 >= p2 && p1 + p2 + 1 < size) {
                if (bits[p1 - p2] != bits[p1 + p2 + 1]) {
                    match = false
                    break
                }
                p2++
            }
            if (match) return p1 + 1
        }
        return 0
    }

    fun smudged(bits: IntArray, size: Int): Int {
        for (p1 in 0 until size - 1) {
            var smudges = 0
            var p2 = 0
            while (p1 >= p2 && p1 + p2 + 1 < size) {
                smudges += (bits[p1 - p2] xor bits[p1 + p2 + 1]).countOneBits()
                if (smudges > 1) break
                p2++
            }
            if (smudges == 1) return p1 + 1
        }
        return 0
    }

    fun part1(input: List<String>): Int {
        val tmpList = ArrayList<String>()
        var result = 0
        for ((idx, i) in input.withIndex()) {
            if (i.isNotBlank()) {
                tmpList.add(i)
            }
            if (i.isBlank() || (idx == input.lastIndex && tmpList.isNotEmpty())) {
                val grid = CharGrid(tmpList)
                val vertSA = IntArray(grid.width)
                val horzSA = IntArray(grid.height)
                for (c in 0 until grid.width) {
                    for (r in 0 until grid.height) {
                        vertSA[c] = vertSA[c] * 2 + (if (grid[c, r] == '#') 1 else 0)
                    }
                }
                for (r in 0 until grid.height) {
                    for (c in 0 until grid.width) {
                        horzSA[r] = horzSA[r] * 2 + if (grid[c, r] == '#') 1 else 0
                    }
                }
                val vref = mirrored(vertSA, grid.width)
                if (vref > 0) {
                    result += vref
                } else {
                    result += mirrored(horzSA, grid.height) * 100
                }

                tmpList.clear()
            }
        }
        return result
    }

    fun part2(input: List<String>): Int {
        val tmpList = ArrayList<String>()
        var result = 0
        for ((idx, i) in input.withIndex()) {
            if (i.isNotBlank()) {
                tmpList.add(i)
            }
            if (i.isBlank() || (idx == input.lastIndex && tmpList.isNotEmpty())) {
                val grid = CharGrid(tmpList)
                val vertSA = IntArray(grid.width)
                val horzSA = IntArray(grid.height)
                for (c in 0 until grid.width) {
                    for (r in 0 until grid.height) {
                        vertSA[c] = vertSA[c] * 2 + (if (grid[c, r] == '#') 1 else 0)
                    }
                }
                for (r in 0 until grid.height) {
                    for (c in 0 until grid.width) {
                        horzSA[r] = horzSA[r] * 2 + if (grid[c, r] == '#') 1 else 0
                    }
                }
                val vref = smudged(vertSA, grid.width)
                if (vref > 0) {
                    result += vref
                } else {
                    result += smudged(horzSA, grid.height) * 100
                }

                tmpList.clear()
            }
        }
        return result
    }

    // test if implementation meets criteria from the description, like:
    val testInput = inlineTestInput.trim().reader().readLines()
    //val testInput = readInput("aoc2023/Day13_test")
    val testInputPart1Result = part1(testInput)
    println("Part 1 Test: $testInputPart1Result")
    val testInputPart2Result = part2(testInput)
    println("Part 2 Test: $testInputPart2Result")
    check(testInputPart1Result == 405)
    check(testInputPart2Result == 400)

    val input = readInput("aoc2023/Day13")
    part1(input).println()
    part2(input).println()
}
