package aoc2024

import CharGrid
import RelPos
import println
import readInput

/*
--- Day 6: Guard Gallivant ---
The Historians use their fancy device again, this time to whisk you all away to the North Pole prototype suit manufacturing lab... in the year 1518! It turns out that having direct access to history is very convenient for a group of historians.
You still have to be careful of time paradoxes, and so it will be important to avoid anyone from 1518 while The Historians search for the Chief. Unfortunately, a single guard is patrolling this part of the lab.
Maybe you can work out where the guard will go ahead of time so that The Historians can search safely?
You start by making a map (your puzzle input) of the situation. For example:
....#.....
.........#
..........
..#.......
.......#..
..........
.#..^.....
........#.
#.........
......#...

The map shows the current position of the guard with ^ (to indicate the guard is currently facing up from the perspective of the map). Any obstructions - crates, desks, alchemical reactors, etc. - are shown as #.
Lab guards in 1518 follow a very strict patrol protocol which involves repeatedly following these steps:

If there is something directly in front of you, turn right 90 degrees.
Otherwise, take a step forward.

Following the above protocol, the guard moves up several times until she reaches an obstacle (in this case, a pile of failed suit prototypes):
....#.....
....^....#
..........
..#.......
.......#..
..........
.#........
........#.
#.........
......#...

Because there is now an obstacle in front of the guard, she turns right before continuing straight in her new facing direction:
....#.....
........>#
..........
..#.......
.......#..
..........
.#........
........#.
#.........
......#...

Reaching another obstacle (a spool of several very long polymers), she turns right again and continues downward:
....#.....
.........#
..........
..#.......
.......#..
..........
.#......v.
........#.
#.........
......#...

This process continues for a while, but the guard eventually leaves the mapped area (after walking past a tank of universal solvent):
....#.....
.........#
..........
..#.......
.......#..
..........
.#........
........#.
#.........
......#v..

By predicting the guard's route, you can determine which specific positions in the lab will be in the patrol path. Including the guard's starting position, the positions visited by the guard before leaving the area are marked with an X:
....#.....
....XXXXX#
....X...X.
..#.X...X.
..XXXXX#X.
..X.X.X.X.
.#XXXXXXX.
.XXXXXXX#.
#XXXXXXX..
......#X..

In this example, the guard will visit 41 distinct positions on your map.
Predict the path of the guard. How many distinct positions will the guard visit before leaving the mapped area?

*/
fun main() {

    val inlineTestInput = """
....#.....
.........#
..........
..#.......
.......#..
..........
.#..^.....
........#.
#.........
......#...
"""

    fun part1(input: List<String>): Int {
        val grid = CharGrid(input, '.')
        val (oc, or) = grid.findMatches { it == '^' }[0]

        var pos = RelPos(oc, or)
        var dir = RelPos(0, -1)
        while (grid.isInside(pos)) {
            grid[pos] = 'X'
            while (grid[pos.translate(dir)] == '#') {
                dir = RelPos(-dir.dr, dir.dc)
            }
            pos = pos.translate(dir)
        }
        return grid.findMatches { it == 'X' }.count()
    }

    fun checkLooping(grid: CharGrid, origPos: RelPos): Boolean {
        val dirgrid = Array(grid.height) { IntArray(grid.width) }
        var dir = RelPos(0, -1)
        var dirBit = 0
        var looping = false
        var pos = origPos
        while (grid.isInside(pos)) {
            if ((dirgrid[pos.dc][pos.dr] and (1 shl dirBit)) != 0) {
                looping = true
                break
            }
            dirgrid[pos.dc][pos.dr] = dirgrid[pos.dc][pos.dr] or (1 shl dirBit)
            while (grid[pos.translate(dir)] == '#') {
                dir = RelPos(-dir.dr, dir.dc)
                dirBit = (dirBit + 1) and 3
            }
            pos = pos.translate(dir)
        }
        return looping
    }

    fun part2(input: List<String>): Int {
        val grid = CharGrid(input, '.')
        val obstaclePos = HashSet<RelPos>()
        val solutions = HashSet<RelPos>()
        val (oc, or) = grid.findMatches { it == '^' }[0]

        var pos = RelPos(oc, or)
        var dir = RelPos(0, -1)
        while (true) {
            grid[pos] = 'X'
            val newPos = pos.translate(dir)
            if (!grid.isInside(newPos)) break
            if (grid[newPos] == '#') {
                dir = RelPos(-dir.dr, dir.dc)
            } else {
                if (grid[newPos] == '.') {
                    obstaclePos.add(newPos)
                }
                pos = newPos
            }
        }
        for (op in obstaclePos) {
            grid[op] = '#'
            if (checkLooping(grid, RelPos(oc, or))) {
                solutions.add(op)
            }
            grid[op] = '.'
        }
        return solutions.size
    }

    // test if implementation meets criteria from the description, like:
    val testInput = inlineTestInput.trim().reader().readLines()
    //val testInput = readInput("aoc2024/Day06_test")
    val testInputPart1Result = part1(testInput)
    println("Part 1 Test: $testInputPart1Result")
    val testInputPart2Result = part2(testInput)
    println("Part 2 Test: $testInputPart2Result")
    check(testInputPart1Result == 41)
    check(testInputPart2Result == 6)

    val input = readInput("aoc2024/Day06")
    part1(input).println()
    part2(input).println()
}
